package fr.jelora.minitel;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.concurrent.Semaphore;

import javax.imageio.ImageIO;

import org.apache.log4j.Logger;

import fr.jelora.jraspitools.RasPiPin;
import fr.jelora.jraspitools.gpio.GPIOException;
import fr.jelora.jraspitools.gpio.GPIOTools;
import fr.jelora.minitel.enums.CodeCaractereSemiGraphique;
import fr.jelora.minitel.enums.CodeCouleurCaractere;
import fr.jelora.minitel.enums.CodeCouleurDeFond;
import fr.jelora.minitel.enums.CodeSpecifique;
import fr.jelora.minitel.enums.ModeEcran;
import fr.jelora.minitel.enums.TailleTexte;
import fr.jelora.minitel.enums.ToucheClavier;
import fr.jelora.minitel.exception.ConnexionException;
import fr.jelora.minitel.exception.ReceptionException;
import fr.jelora.minitel.exception.TransmissionException;
import jssc.SerialPort;
import jssc.SerialPortException;

public class MinitelTools {
	
	final static Logger logger = Logger.getLogger(MinitelTools.class);
	
	private SerialPort serialTransmission1200;
	private SerialPort serialReception75;
	
	private Semaphore mutexBooleanThreadSurveillance;
	private boolean continuerThreadSurveillance;
	
	private Semaphore mutexCompteurActivite;
	private int compteurActivite;
	private Thread threadSurveillanceActivite;
	
	private Thread threadSurveillanceConnexionMinitel;
	
	public MinitelTools(SerialPort serialTransmission1200, SerialPort serialReception75){
		this.serialTransmission1200 = serialTransmission1200;
		this.serialReception75 = serialReception75;
		
		compteurActivite = -1;
		mutexCompteurActivite = new Semaphore(1);
		mutexBooleanThreadSurveillance = new Semaphore(1);					
	}
	
	public void demarrerThreadsSurveillance(RasPiPin pinDecrochageLigne, RasPiPin pinActivationLiaisonDonnees, RasPiPin pinDectectionConnexionMinitel, boolean modeDev) throws InterruptedException{
		mutexBooleanThreadSurveillance.acquire();
		continuerThreadSurveillance = true;
		mutexBooleanThreadSurveillance.release();
		
		// Thread de déconnexion automatique en cas de non activité au bout de
		// 10 minutes
		threadSurveillanceActivite = new Thread("ThreadSurveillanceActivite") {
			public void run() {
				GPIOTools gpioTools = new GPIOTools();

				while (true) {
					try {
						mutexBooleanThreadSurveillance.acquire();
						if (continuerThreadSurveillance) {
							mutexBooleanThreadSurveillance.release();

							if (serialTransmission1200.isOpened() && serialReception75.isOpened()) {
								mutexCompteurActivite.acquire();
								if (compteurActivite == 0) {
									// Coupure connexion
									logger.info("Pas d'activitée depuis 10 minutes. Déconnexion.");
									serialReception75.closePort();
									serialTransmission1200.closePort();
									if (!modeDev) {
										gpioTools.setValue(pinActivationLiaisonDonnees, false);
										gpioTools.setValue(pinDecrochageLigne, false);
									}
								}
								if (compteurActivite >= 0) {
									compteurActivite--;
								}
								mutexCompteurActivite.release();
							} 
							else {
								break;
							}
						} 
						else {
							mutexBooleanThreadSurveillance.release();
							break;
						}
						Thread.sleep(100);
					} 
					catch (InterruptedException e) {
						logger.error(e);
						break;
					} 
					catch (SerialPortException e) {
						logger.error(e);
						break;
					} 
					catch (GPIOException e) {
						logger.error(e);
						break;
					}
				}
			}
		};
		threadSurveillanceActivite.start();
		
		// Thread de raccrochage ligne en cas de perte de liaison Minitel
		threadSurveillanceConnexionMinitel = new Thread("ThreadSurveillanceActivite") {
			public void run() {
				GPIOTools gpioTools = new GPIOTools();
				int comptageVerificationConnexion = 0;
				while (true) {
					try {
						if (!modeDev) {
							mutexBooleanThreadSurveillance.acquire();
							if (continuerThreadSurveillance) {
								mutexBooleanThreadSurveillance.release();
								if (serialTransmission1200.isOpened() && serialReception75.isOpened() && !gpioTools.getValueBoolean(pinDectectionConnexionMinitel)) {
									comptageVerificationConnexion = 0;
									while ( serialTransmission1200.isOpened() && serialReception75.isOpened() && !gpioTools.getValueBoolean(pinDectectionConnexionMinitel) && continuerThreadSurveillance ){
										comptageVerificationConnexion++;
										if (comptageVerificationConnexion == 30) {
											// Connexion complètement inactive
											// depuis 2 secondes
											// Déconnexion et raccrochage
											logger.info("Coupure liaison Minitel depuis 2 secondes. Déconnexion.");
											serialReception75.closePort();
											serialTransmission1200.closePort();
											if (!modeDev) {
												gpioTools.setValue(pinActivationLiaisonDonnees, false);
												gpioTools.setValue(pinDecrochageLigne, false);
											}
										}
										Thread.sleep(100);
									}
								}
							} 
							else {
								mutexBooleanThreadSurveillance.release();
								break;
							}
						} 
						else {
							break;
						}
						Thread.sleep(100);
					} 
					catch (InterruptedException e) {
						logger.error(e);
						break;
					} 
					catch (SerialPortException e) {
						logger.error(e);
						break;
					} 
					catch (GPIOException e) {
						logger.error(e);
						break;
					}
				}
			}
		};
		threadSurveillanceConnexionMinitel.start();
	}
	
	public void arreterThreadsSurveillance() throws InterruptedException {
		mutexBooleanThreadSurveillance.acquire();
		continuerThreadSurveillance = false;
		mutexBooleanThreadSurveillance.release();
	}
	
	public void signalerActiviteReception() throws ReceptionException{
		try {
			mutexCompteurActivite.acquire();
			compteurActivite = 6000; // 10 minutes
			mutexCompteurActivite.release();
		}
		catch (InterruptedException e) {
			throw new ReceptionException(e);
		}
	}
	
	public void write(byte[] data) throws TransmissionException {
		try {
			if ( serialTransmission1200 != null ){
				if ( serialTransmission1200.isOpened() ){
					serialTransmission1200.writeBytes(data);
					
					// Ralentissement du code pour attendre que le lot de valeurs soit envoyées au Minitel
					Thread.sleep(data.length * 8);
				}
				else{
					throw new ConnexionException("SerialPort fermé");
				}
			}
			else{
				throw new ConnexionException("SerialPort null");
			}
		}
		catch (SerialPortException e) {
			throw new TransmissionException(e);
		} 
		catch (ConnexionException e) {
			throw new TransmissionException(e);
		} 
		catch (InterruptedException e) {
			throw new TransmissionException(e);
		}
	}
	
	public byte readByte() throws ReceptionException {
		byte byteReaded = -1;
		try {
			if ( serialReception75 != null ){
				while ( byteReaded == -1 ){
					if ( serialReception75.isOpened() ){
						byte[] bytes = serialReception75.readBytes();
						if ( bytes != null ){
							if ( bytes.length > 0 ){
								byteReaded = bytes[bytes.length - 1];
								signalerActiviteReception();
							}
						}
					}
					else{
						throw new ConnexionException("SerialPort fermé");
					}
					Thread.sleep(50);
				}
			}
			else{
				throw new ConnexionException("SerialPort null");
			}
		}
		catch (SerialPortException e) {
			throw new ReceptionException(e);
		}
		catch (ConnexionException e) {
			throw new ReceptionException(e);
		} 
		catch (InterruptedException e) {
			throw new ReceptionException(e);
		}
		return byteReaded;
	}
	
	public ToucheClavier readClavier() throws ReceptionException {
		byte byteLu = this.readByte();
		if ( byteLu == CodeSpecifique.Sep.toByte() || byteLu == CodeSpecifique.SS2.toByte() ){
			// Touche spécifique 2 octects
			// Attente 2e octect
			byte byteLu2 = this.readByte();
			for ( ToucheClavier toucheClavier : ToucheClavier.values() ){
				if ( new String(new byte[] { byteLu, byteLu2 }).equals(toucheClavier.toString()) ){
					return toucheClavier;
				}
			}
			// Aucune touche clavier correspondante
			return null;
		}
		else if ( byteLu == CodeSpecifique.ESC.toByte() ){
			// Touche spécifique 3 octects
			// Attente 2e et 3e octect
			byte byteLu2 = this.readByte();
			byte byteLu3 = this.readByte();
			for ( ToucheClavier toucheClavier : ToucheClavier.values() ){
				if ( new String(new byte[] { byteLu, byteLu2, byteLu3 }).equals(toucheClavier.toString()) ){
					return toucheClavier;
				}
			}
			// Aucune touche clavier correspondante
			return null;
		}
		else{
			// Touche normale
			for ( ToucheClavier toucheClavier : ToucheClavier.values() ){
				if ( new String(new byte[] { byteLu }).equals(toucheClavier.toString()) ){
					return toucheClavier;
				}
			}
			// Aucune touche clavier correspondante
			return null;
		}
		
	}
	
	public void viderCacheRead() throws ReceptionException {
		try {
			if ( serialReception75 != null ){
				if ( serialReception75.isOpened() ){
					serialReception75.readBytes();
					signalerActiviteReception();
				}
				else{
					throw new ConnexionException("SerialPort fermé");
				}
			}
			else{
				throw new ConnexionException("SerialPort null");
			}
		}
		catch (SerialPortException e) {
			throw new ReceptionException(e);
		}
		catch (ConnexionException e) {
			throw new ReceptionException(e);
		}
	}
	
	// Utilisé pour le débug
	/*private String bytesToHex(byte[] data){
		StringBuffer hexString = new StringBuffer();
		for (int i = 0; i < data.length; i++) {
			String hex = Integer.toHexString(0xff & data[i]);
			if (hex.length() == 1) {
				hexString.append('0');
			}
			hexString.append(hex);
		}
		return hexString.toString();
	}*/
	
	public void write(String data) throws TransmissionException{		
		this.write(this.toStringMinitel(data).getBytes());
	}
	
	private String toStringMinitel(String data){
		data = data.replace("â", ToucheClavier.AccentCirconflexe.toString() + "a");
		data = data.replace("á", ToucheClavier.AccentAigu.toString() + "a");
		data = data.replace("à", ToucheClavier.AccentGrave.toString() + "a");
		data = data.replace("ä", ToucheClavier.AccentTrema.toString() + "a");
		data = data.replace("ê", ToucheClavier.AccentCirconflexe.toString() + "e");
		data = data.replace("ë", ToucheClavier.AccentTrema.toString() + "e");
		data = data.replace("é", ToucheClavier.AccentAigu.toString() + "e");
		data = data.replace("è", ToucheClavier.AccentGrave.toString() + "e");
		data = data.replace("î", ToucheClavier.AccentCirconflexe.toString() + "i");
		data = data.replace("ï", ToucheClavier.AccentTrema.toString() + "i");
		data = data.replace("ì", ToucheClavier.AccentGrave.toString() + "i");
		data = data.replace("í", ToucheClavier.AccentAigu.toString() + "i");
		data = data.replace("ô", ToucheClavier.AccentCirconflexe.toString() + "o");
		data = data.replace("ö", ToucheClavier.AccentTrema.toString() + "o");
		data = data.replace("ò", ToucheClavier.AccentGrave.toString() + "o");
		data = data.replace("ó", ToucheClavier.AccentAigu.toString() + "o");
		data = data.replace("û", ToucheClavier.AccentCirconflexe.toString() + "u");
		data = data.replace("ü", ToucheClavier.AccentTrema.toString() + "u");
		data = data.replace("ù", ToucheClavier.AccentGrave.toString() + "u");
		data = data.replace("ú", ToucheClavier.AccentAigu.toString() + "u");
		data = data.replace("ç", ToucheClavier.Cedille.toString() + "c");
		data = data.replace("À", "A");
		data = data.replace("Á", "A");
		data = data.replace("Â", "A");
		data = data.replace("Ä", "A");
		data = data.replace("È", "E");
		data = data.replace("É", "E");
		data = data.replace("Ê", "E");
		data = data.replace("Ë", "E");
		data = data.replace("Ì", "I");
		data = data.replace("Í", "I");
		data = data.replace("Î", "I");
		data = data.replace("Ï", "I");
		data = data.replace("Ò", "O");
		data = data.replace("Ó", "O");
		data = data.replace("Ô", "O");
		data = data.replace("Ö", "O");
		data = data.replace("Ù", "U");
		data = data.replace("Ú", "U");
		data = data.replace("Û", "U");
		data = data.replace("Ü", "U");
		data = data.replace("Ç", "C");
		return data;
	}
	
	public void clearScreen() throws TransmissionException {
		this.write("\f");
	}
	
	public void clearLigneEnCours() throws TransmissionException {
		this.write(CodeSpecifique.CSI.toByteArray());
		this.write("2K");
	}
	public void clearLigneDuCurseurALaFinDeLaLigne() throws TransmissionException {
		this.write(CodeSpecifique.CSI.toByteArray());
		this.write("K");
	}
	
	public void clearLigne00() throws TransmissionException{
		this.allerDansLigne00(1);
		this.writeRepeteByte((byte) ' ', 39);
		this.write("\n");
	}
	
	public void clearAll() throws TransmissionException {
		this.setCursorBlinking(false);
		this.clearLigne00();
		this.clearScreen();
	}
	
	public void allerDansLigne00(int colonne) throws TransmissionException{
		byte colonneCode = (byte) (0x40 + colonne);
		this.write(new byte[]{ 0x1F, 0x40, colonneCode });
	}
	
	public void writeRepeteByte(byte byteRepetition, int nbRepetitions) throws TransmissionException{
        byte nbRepetitionsCode = (byte) (0x40 + nbRepetitions);
		this.write(new byte[] { byteRepetition, 0x12, nbRepetitionsCode});
	}
	
	public void locate(int posX, int posY) throws TransmissionException{
		if ( posX >= 1 && posX <= 40 && posY >= 1 && posY <= 24 ){
			this.write(CodeSpecifique.CSI.toString());
	        
	        byte dixaines = (byte) Math.round(posY / 10);
	        byte unite = (byte) (posY- Math.round(posY / 10) * 10);
	        byte dixinesCode = (byte) (0x30 + dixaines);
	        byte unitesCode = (byte) (0x30 + unite);
	        
	        this.write(new byte[] { dixinesCode, unitesCode, ';' });
	        
	        dixaines = (byte) Math.round(posX / 10);
	        unite = (byte) (posX - Math.round(posX / 10) * 10);
	        
	        dixinesCode = (byte) (0x30 + dixaines);
	        unitesCode = (byte) (0x30 + unite);
	        
	        this.write(new byte[] { dixinesCode, unitesCode, 'H' });
		}
	}
	
	public void setColorText(CodeCouleurCaractere colorText) throws TransmissionException{
		this.write(new String(new byte[] { 0x1B }));
		this.write(colorText.toString());
	}
	
	public void setColorBackground(CodeCouleurDeFond colorBackground) throws TransmissionException{
		this.write(new String(new byte[] { 0x1B }));
		this.write(colorBackground.toString());
	}
	
	public void setTextClignotant(boolean texteClignotant) throws TransmissionException{
		this.write(new String(new byte[] { 0x1B }));
		if ( texteClignotant ){
			this.write(new String(new byte[] { 0x48 }));
		}
		else{
			this.write(new String(new byte[] { 0x49 }));
		}
	}
	
	public String optimiserTexte(String data){
		String optimiseText = "";
		
		int nbRepetitions;
		String carac;
		int posTemp;
		boolean stop = false;
		for ( int i = 0 ; i < data.length() ; i++ ){
			
			nbRepetitions = 0;
			carac = data.substring(i, i+1);
			posTemp = i+1;
			stop = false;
			while ( !stop ){
				if ( posTemp < data.length() ){
					if ( carac.equals(data.substring(posTemp, posTemp+1)) ){
						nbRepetitions++;
					}
					else{
						stop = true;
					}
				}
				else{
					stop = true;
				}
				posTemp++;
			}
			
			optimiseText += carac;
			
			if ( nbRepetitions > 3 ){
				byte nbRepeatCode = (byte) (0x40 + nbRepetitions);
				optimiseText += new String(new byte[] { 0x12, nbRepeatCode});
			}
			else{
				for ( int repet = 0 ; repet < nbRepetitions ; repet++ ){
					optimiseText += carac;
				}
			}
			
			i = i + nbRepetitions;
		}
		
		return optimiseText;
	}
	
	public void setTailleTexte(TailleTexte textSize) throws TransmissionException{
        this.write(CodeSpecifique.ESC.toByteArray());
        this.write(textSize.toString());
        this.write(new byte[] { 0x20 });
	}
	
	public void setCursorBlinking(boolean blinkingStatut) throws TransmissionException{
		if ( blinkingStatut ){
			this.write(CodeSpecifique.Con.toByteArray());
		}
		else{
			this.write(CodeSpecifique.Coff.toByteArray());
		}
	}
	
	public void writeImage(int posX, int posY, File input) throws TransmissionException{
		try{
			BufferedImage image = ImageIO.read(input);
	        int width = image.getWidth();
	        if ( width > 78 - ((posX - 1) * 2) ) width = 78 - ((posX - 1) * 2);
	        int height = image.getHeight();
	        if ( height > 72 - ((posY - 1)* 3) ) height = 72 - ((posY - 1)* 3);
	        
	        int nbCaracHeight = (int) Math.ceil(height / 3);
	        int nbCaracWidth = (int) Math.ceil(width / 2);
	        
	        int colorMin;
	        int colorMax;
	        int colorNiveauxGris;
	        Color colorCarac;
	        String codeCarac = "";
	        
	        int colorFondPreced = -1;
	        int colorTextPreced = -1;
	        String tempCaracs = "";
	        
	        this.write(CodeSpecifique.SO.toString());
	        for ( int caracHeight = 0 ; caracHeight < nbCaracHeight ; caracHeight++ ){
	        	
	        	this.locate(posX, posY + caracHeight);
	        	
		        for ( int caracWidth = 0 ; caracWidth < nbCaracWidth ; caracWidth++ ){
		        	
		        	// Recherche niveau de gris min et max
		        	colorMin = 7;
		        	colorMax = 0;
		        	for ( int y = 0 ; y<3 ; y++ ){
		    			for ( int x = 0 ; x<2 ; x++ ){
		    				colorCarac = new Color(image.getRGB(x+caracWidth*2, y+caracHeight*3));
		    				colorNiveauxGris = (colorCarac.getRed() + colorCarac.getGreen() + colorCarac.getBlue()) / 96;
		    				if ( colorNiveauxGris > colorMax ){
		    					colorMax = colorNiveauxGris;
		    				}
		    				if ( colorNiveauxGris < colorMin ){
		    					colorMin = colorNiveauxGris;
		    				}
		    			}
		    		}
		        	
		        	// Séparation en 2 couleurs
		        	codeCarac = "";
		        	for ( int y = 0 ; y<3 ; y++ ){
		    			for ( int x = 0 ; x<2 ; x++ ){
		    				colorCarac = new Color(image.getRGB(x+caracWidth*2, y+caracHeight*3));
		    				colorNiveauxGris = (colorCarac.getRed() + colorCarac.getGreen() + colorCarac.getBlue()) / 96;
		    				if ( colorNiveauxGris == colorMin && colorNiveauxGris == colorMax ){
		    					codeCarac += "1";
		    				}
		    				else if ( colorNiveauxGris == colorMin ){
		    					codeCarac += "0";
		    				}
		    				else if ( colorNiveauxGris == colorMax ){
		    					codeCarac += "1";
		    				}
		    				else{
		    					if ( (colorNiveauxGris - colorMin) < ((colorMax - colorMin)/2) ){
		    						codeCarac += "0";
		    					}
		    					else{
		    						codeCarac += "1";
		    					}
		    				}
		    			}
		        	}
		        	
		        	// Changement possible de couleur de fond et/ou de texte
		        	if ( colorMin != colorFondPreced || colorMax != colorTextPreced){
		        		
		        		tempCaracs = this.optimiserTexte(tempCaracs);
		        		this.write(tempCaracs);
		        		tempCaracs = "";
		        		
				        // Couleur de fond
				        if ( colorMin != colorFondPreced ){
					       	this.setColorBackground(CodeCouleurDeFond.valueOf("fondGris" + colorMin));
					       	colorFondPreced = colorMin;
				        }
				        	
						// Couleur caractere
						if (colorMax != colorTextPreced) {
							this.setColorText(CodeCouleurCaractere.valueOf("caractereGris" + colorMax));
							colorTextPreced = colorMax;
						}
		        	}
		        	
		        	tempCaracs += CodeCaractereSemiGraphique.valueOf("code" + codeCarac).toString();
		        }
		        
		        tempCaracs = this.optimiserTexte(tempCaracs);
		        this.write(tempCaracs);
		        tempCaracs = "";
		        
		        colorFondPreced = -1;
		        colorTextPreced = -1;
		        
	        }
	        this.write(CodeSpecifique.SI.toString());
		}
		catch( IOException e ){
			throw new TransmissionException(e);
		} 
	}
	
	public void faireUnBip() throws TransmissionException{
		this.write(new byte[]{ 0x07 });
	}
	
	public void changerModeEcran(ModeEcran modeEcran) throws TransmissionException, ReceptionException{
		if ( modeEcran != null ){
			CodeSpecifique[] codesSpecifiques = modeEcran.getCodesSpecifiques();
			for ( int i = 0 ; i < codesSpecifiques.length ; i++ ){
				this.write(codesSpecifiques[i].toByteArray());
			}
		}
	}
}
